# Copyright 2021 Sushil Khanchi <sushilkhanchi97@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest
import grpc
import re
from hgitaly.git import (
    EMPTY_TREE_OID,
    NULL_BLOB_OID,
)
from hgitaly.oid import tree_oid
from hgitaly.revision import (
    gitlab_revision_hash,
)
from hgitaly.stub.commit_pb2 import (
    TreeEntryRequest,
    TreeEntryResponse,
)
from hgitaly.stub.diff_pb2 import (
    CommitDeltaRequest,
    CommitDiffRequest,
    DiffStatsRequest,
    FindChangedPathsRequest,
    GetPatchIDRequest,
    RawDiffRequest,
    RawPatchRequest,
)
from hgitaly.stub.commit_pb2_grpc import CommitServiceStub
from hgitaly.stub.diff_pb2_grpc import DiffServiceStub
from .test_blob_tree import oid_mapping  # TODO make it a fixture method

from . import skip_comparison_tests
if skip_comparison_tests():  # pragma no cover
    pytestmark = pytest.mark.skip

INDEX_LINE_REGEXP = re.compile(br'^index \w+\.\.\w+( \d+)?$')


def test_compare_raw_diff(gitaly_comparison):
    fixture = gitaly_comparison
    wrapper_repo = fixture.gitaly_repo
    git_repo = fixture.git_repo
    wrapper = fixture.hg_repo_wrapper

    gl_branch = b'branch/default'
    ctx0 = wrapper.commit_file('bar', content="I am in\nrab\n",
                               message="Some bar")
    git_shas = {
        ctx0.hex(): git_repo.branches()[gl_branch]['sha']
    }
    ctx1 = wrapper.commit_file('bar', content="I am in\nbar\n",
                               message="Changes bar")
    git_shas[ctx1.hex()] = git_repo.branches()[gl_branch]['sha']
    ctx2 = wrapper.commit_file('zoo', content="I am in\nzoo\n",
                               message="Added zoo")
    git_shas[ctx2.hex()] = git_repo.branches()[gl_branch]['sha']
    wrapper.command(b'mv', wrapper.repo.root + b'/bar',
                    wrapper.repo.root + b'/zar')
    wrapper.command(b'ci', message=b"Rename bar to zar")
    ctx3 = wrapper.repo[b'.']
    git_shas[ctx3.hex()] = git_repo.branches()[gl_branch]['sha']
    # Repo structure:
    #
    # @  3 Rename bar to zar
    # |
    # o  2 Added zoo
    # |
    # o  1 Changes bar
    # |
    # o  0 Some bar
    #

    diff_stubs = dict(
        git=DiffServiceStub(fixture.gitaly_channel),
        hg=DiffServiceStub(fixture.hgitaly_channel)
    )

    def do_rpc(vcs, left_cid, right_cid):
        if vcs == 'git':
            left_cid = git_shas.get(left_cid, left_cid)
            right_cid = git_shas.get(right_cid, right_cid)
        request = RawDiffRequest(
                            repository=wrapper_repo,
                            left_commit_id=left_cid,
                            right_commit_id=right_cid,
                        )
        response = diff_stubs[vcs].RawDiff(request)
        return b''.join(resp.data for resp in response)

    # case 1: when indexline doesn't contain <mode>
    hg_resp_lines = do_rpc('hg', ctx1.hex(), ctx2.hex()).split(b'\n')
    git_resp_lines = do_rpc('git', ctx1.hex(), ctx2.hex()).split(b'\n')
    INDEX_LINE_POSITION = 2
    hg_indexline = hg_resp_lines[INDEX_LINE_POSITION]
    git_indexline = git_resp_lines[INDEX_LINE_POSITION]
    # check that index line has the correct format
    assert INDEX_LINE_REGEXP.match(hg_indexline) is not None
    assert INDEX_LINE_REGEXP.match(git_indexline) is not None
    # actual comparison
    del hg_resp_lines[INDEX_LINE_POSITION]
    del git_resp_lines[INDEX_LINE_POSITION]
    assert hg_resp_lines == git_resp_lines

    # case 2: when indexline has <mode> (it happens when mode didn't change)
    hg_resp_lines = do_rpc('hg', ctx0.hex(), ctx1.hex()).split(b'\n')
    git_resp_lines = do_rpc('git', ctx0.hex(), ctx1.hex()).split(b'\n')
    INDEX_LINE_POSITION = 1
    hg_indexline = hg_resp_lines[INDEX_LINE_POSITION]
    git_indexline = git_resp_lines[INDEX_LINE_POSITION]
    # check the mode
    assert INDEX_LINE_REGEXP.match(hg_indexline).group(1) == b' 100644'
    assert INDEX_LINE_REGEXP.match(git_indexline).group(1) == b' 100644'

    # case 3: test with file renaming
    hg_resp = do_rpc('hg', ctx2.hex(), ctx3.hex())
    git_resp = do_rpc('git', ctx2.hex(), ctx3.hex())
    assert hg_resp is not None
    assert hg_resp == git_resp

    # case 4: when commit_id does not correspond to a commit
    sha_not_exists = b'deadnode' * 5
    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc('hg', sha_not_exists, ctx2.hex())
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc('git', sha_not_exists, ctx2.hex())
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL

    # case 5: EMPTY_TREE_OID to represent null commit on the left
    hg_resp_lines = do_rpc('hg', EMPTY_TREE_OID, ctx0.hex()).split(b'\n')
    git_resp_lines = do_rpc('git', EMPTY_TREE_OID, ctx0.hex()).split(b'\n')
    INDEX_LINE_POSITION = 2
    assert (hg_resp_lines[INDEX_LINE_POSITION].split(b'..')[0]
            ==
            git_resp_lines[INDEX_LINE_POSITION].split(b'..')[0])
    del hg_resp_lines[INDEX_LINE_POSITION]
    del git_resp_lines[INDEX_LINE_POSITION]
    assert hg_resp_lines == git_resp_lines

    # case 6: EMPTY_TREE_OID to represent null commit on the right
    git_resp_lines = do_rpc('git', ctx0.hex(), EMPTY_TREE_OID).split(b'\n')
    hg_resp_lines = do_rpc('hg', ctx0.hex(), EMPTY_TREE_OID).split(b'\n')
    INDEX_LINE_POSITION = 2
    assert (hg_resp_lines[INDEX_LINE_POSITION].rsplit(b'..')[-1]
            ==
            git_resp_lines[INDEX_LINE_POSITION].rsplit(b'..')[-1])
    del hg_resp_lines[INDEX_LINE_POSITION]
    del git_resp_lines[INDEX_LINE_POSITION]
    assert hg_resp_lines == git_resp_lines


def test_compare_raw_patch(gitaly_comparison):
    fixture = gitaly_comparison
    gitaly_repo = fixture.gitaly_repo
    git_repo = fixture.git_repo
    wrapper = fixture.hg_repo_wrapper

    gl_branch = b'branch/default'
    ctx0 = wrapper.commit_file('bar', content="I am in\nrab\n",
                               message="Some bar")
    git_sha0 = git_repo.branches()[gl_branch]['sha']

    diff_stubs = dict(
        git=DiffServiceStub(fixture.gitaly_channel),
        hg=DiffServiceStub(fixture.hgitaly_channel)
    )

    def do_rpc(vcs, left_cid, right_cid):
        request = RawPatchRequest(
            repository=gitaly_repo,
            left_commit_id=left_cid,
            right_commit_id=right_cid,
        )
        response = diff_stubs[vcs].RawPatch(request)
        return b''.join(resp.data for resp in response)

    # Here we are only comparing for error cases, as HGitaly returns Hg patches
    # and Gitaly returns Git pataches. For more, look at diff.RawPatch()
    sha_not_exists = b'deadnode' * 5
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        do_rpc('hg', sha_not_exists, ctx0.hex())
    with pytest.raises(grpc.RpcError) as exc_info_git:
        do_rpc('git', sha_not_exists, git_sha0)
    assert exc_info_hg.value.code() == exc_info_git.value.code()
    assert exc_info_hg.value.details() == exc_info_git.value.details()


def test_compare_commit_diff(gitaly_comparison):
    fixture = gitaly_comparison
    gitaly_repo = fixture.gitaly_repo
    git_repo = fixture.git_repo
    wrapper = fixture.hg_repo_wrapper
    gl_branch = b'branch/default'

    wrapper.commit_file('bar', content='I am in\nrab\n',
                        message='Add bar')
    ctx1 = wrapper.commit_file('bar', content='I am in\nbar\n',
                               message='Changes bar')
    git_sha1 = git_repo.branches()[gl_branch]['sha']
    wrapper.command(b'mv', wrapper.repo.root + b'/bar',
                    wrapper.repo.root + b'/zar')
    wrapper.command(b'ci', message=b'Rename bar to zar')
    ctx3 = wrapper.commit_file('zoo', content='I am in \n zoo\n',
                               message='Added zoo')
    git_sha3 = git_repo.branches()[gl_branch]['sha']
    ctx4 = wrapper.commit_file('zoo', content='I am in\nzoo\n',
                               message='zoo: change whitespace')
    git_sha4 = git_repo.branches()[gl_branch]['sha']
    ctx5 = wrapper.commit_file('too', content='I am in\ntoo\n',
                               message='Added too')
    git_sha5 = git_repo.branches()[gl_branch]['sha']
    ctx6 = wrapper.commit_file('hulk', content=b'abcd\n' * 1024 + b'foobar\n',
                               message='Added hulk')
    git_sha6 = git_repo.branches()[gl_branch]['sha']
    # Repo structure:
    #
    # @  5 Added too
    # |
    # o  4 zoo: remove whitespace
    # |
    # o  3 Added zoo
    # |
    # o  2 Rename bar to zar
    # |
    # o  1 Changes bar
    # |
    # o  0 Add bar
    #

    def do_rpc(vcs, left_cid, right_cid, **opts):
        request = CommitDiffRequest(
                    repository=gitaly_repo,
                    left_commit_id=left_cid,
                    right_commit_id=right_cid,
                    **opts,
                )
        diff_stubs = dict(git=DiffServiceStub(fixture.gitaly_channel),
                          hg=DiffServiceStub(fixture.hgitaly_channel))
        response = diff_stubs[vcs].CommitDiff(request)
        final = []
        # `from_id`, `to_id` are different in hgitaly/gitaly responses
        # for equality setting them to empty string
        for resp in response:
            resp.from_id = ''
            resp.to_id = ''
            final.append(resp)
        return final

    def rpc_diff_size(*args, **kwargs):
        """Perform RPC and return the total length of patch data it gives."""
        return sum(len(chunk.raw_patch_data)
                   for chunk in do_rpc(*args, **kwargs))

    # case 1: actual test (rename + new file + content change)
    hg_resp = do_rpc('hg', left_cid=ctx1.hex(),
                     right_cid=ctx3.hex())
    git_resp = do_rpc('git', left_cid=git_sha1,
                      right_cid=git_sha3)
    assert hg_resp == git_resp

    # case 2: test with enforce_limits opt
    # thresholds dict with limits checked manually
    thresholds = dict(
        max_files=3,
        max_bytes=29,
        max_lines=5,
        max_patch_bytes=31,
    )
    for lm in [None, 'max_files', 'max_bytes', 'max_lines', 'max_patch_bytes']:
        new_thresholds = thresholds.copy()
        if lm is not None:
            new_thresholds[lm] -= 1

        hg_resp = do_rpc(
            'hg', left_cid=ctx1.hex(),
            right_cid=ctx5.hex(),
            enforce_limits=True,
            **new_thresholds,
        )
        git_resp = do_rpc(
            'git', left_cid=git_sha1,
            right_cid=git_sha5,
            enforce_limits=True,
            **new_thresholds,
        )
        assert hg_resp == git_resp

    # case 3: test with collapse_diffs opt
    # thresholds dict with limits checked manually
    thresholds = dict(
        safe_max_files=3,
        safe_max_bytes=29,
        safe_max_lines=5,
    )
    for lm in [None, 'safe_max_files', 'safe_max_bytes', 'safe_max_lines']:
        new_thresholds = thresholds.copy()
        if lm is not None:
            new_thresholds[lm] -= 1
        hg_resp = do_rpc(
            'hg', left_cid=ctx1.hex(),
            right_cid=ctx5.hex(),
            collapse_diffs=True,
            **new_thresholds,
        )
        git_resp = do_rpc(
            'git', left_cid=git_sha1,
            right_cid=git_sha5,
            collapse_diffs=True,
            **new_thresholds,
        )
        assert hg_resp == git_resp

    # case 4: when commit_id does not correspond to a commit
    sha_not_exists = b'deadnode' * 5
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        do_rpc('hg', sha_not_exists, ctx3.hex())
    with pytest.raises(grpc.RpcError) as exc_info_git:
        do_rpc('git', sha_not_exists, git_sha3)
    assert exc_info_hg.value.code() == exc_info_git.value.code()
    assert exc_info_hg.value.details() == exc_info_git.value.details()

    # case 5: when response.raw_patch_data size exceeds DIFF_MSG_SIZE_THRESHOLD
    hg_resp = do_rpc('hg', left_cid=ctx5.hex(),
                     right_cid=ctx6.hex())
    git_resp = do_rpc('git', left_cid=git_sha5,
                      right_cid=git_sha6)
    assert hg_resp == git_resp
    assert len(hg_resp) == 2

    # case 6: whitespace options
    WHITESPACE_IGNORE = CommitDiffRequest.WHITESPACE_CHANGES_IGNORE
    WHITESPACE_IGNORE_ALL = CommitDiffRequest.WHITESPACE_CHANGES_IGNORE_ALL
    hg_resp = do_rpc('hg', left_cid=ctx3.hex(),
                     right_cid=ctx4.hex(),
                     whitespace_changes=WHITESPACE_IGNORE)
    git_resp = do_rpc('git', left_cid=git_sha3,
                      right_cid=git_sha4,
                      whitespace_changes=WHITESPACE_IGNORE)
    # Git and Mercurial differ in how they represent the line with
    # ignored changes. Both mark it as unchanged in the patch, but
    # Git displays the new value for context Mercurial displays the old value:
    hg_raw = hg_resp[0].raw_patch_data.splitlines()
    # we don't change the first char (' ', '-', or '+')
    hg_raw[1] = hg_raw[1][:-1]
    hg_raw.append(b'')  # for final EOL
    hg_resp[0].raw_patch_data = b'\n'.join(hg_raw)
    assert hg_resp == git_resp

    hg_size = rpc_diff_size('hg', left_cid=ctx3.hex(),
                            right_cid=ctx4.hex(),
                            whitespace_changes=WHITESPACE_IGNORE_ALL)
    git_size = rpc_diff_size('git', left_cid=git_sha3,
                             right_cid=git_sha4,
                             whitespace_changes=WHITESPACE_IGNORE_ALL)
    assert hg_size == git_size


def test_compare_commit_delta(gitaly_comparison):
    fixture = gitaly_comparison
    gitaly_repo = fixture.gitaly_repo
    git_repo = fixture.git_repo
    wrapper = fixture.hg_repo_wrapper
    gl_branch = b'branch/default'

    sha0 = wrapper.commit_file('bar', content="I am in\nrab\n",
                               message="Add bar").hex()
    git_shas = {
        sha0: git_repo.branches()[gl_branch]['sha']
    }
    sha1 = wrapper.commit_file('bar', content="I am in\nbar\n",
                               message="Changes bar").hex()
    git_shas[sha1] = git_repo.branches()[gl_branch]['sha']
    wrapper.command(b'mv', wrapper.repo.root + b'/bar',
                    wrapper.repo.root + b'/zar')
    sha2 = wrapper.commit([b'zar', b'bar'], message=b"Rename bar to zar").hex()
    git_shas[sha2] = git_repo.branches()[gl_branch]['sha']
    sha3 = wrapper.commit_file('zoo', content="I am in \nzoo\n",
                               message="Added zoo").hex()
    git_shas[sha3] = git_repo.branches()[gl_branch]['sha']
    # Repo structure:
    #
    # @  3 Added zoo
    # |
    # o  2 Rename bar to zar
    # |
    # o  1 Changes bar
    # |
    # o  0 Add bar
    #

    diff_stubs = dict(git=DiffServiceStub(fixture.gitaly_channel),
                      hg=DiffServiceStub(fixture.hgitaly_channel))

    def do_rpc(vcs, left_cid, right_cid, **opts):
        request = CommitDeltaRequest(
                    repository=gitaly_repo,
                    left_commit_id=left_cid,
                    right_commit_id=right_cid,
                    **opts,
                )
        response = diff_stubs[vcs].CommitDelta(request)
        final = [resp.deltas for resp in response if resp.deltas]
        # `from_id`, `to_id` are different in hgitaly/gitaly responses
        # for equality setting them to empty string
        for deltas in final:
            for delta in deltas:
                delta.from_id = ''
                delta.to_id = ''
        return final

    def do_assert(left_cid, right_cid, paths):
        assert (
            do_rpc('hg', left_cid, right_cid, paths=paths)
            ==
            do_rpc('git', git_shas[left_cid], git_shas[right_cid], paths=paths)
        )

    # actual test
    test_paths = [
        [],
        [b'zoo', b'zar'],
        [b'zoo', b'bar'],
        [b'bar', b'zar'],
        [b'zoo'],
        [b'bar'],
        [b'zar'],
    ]
    for left_cid in [sha0, sha1, sha2, sha3]:
        for right_cid in [sha0, sha1, sha2, sha3]:
            for paths in test_paths:
                do_assert(left_cid, right_cid, paths=paths)

    # when commit_id does not correspond to a commit
    sha_not_exists = b'deadnode' * 5
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        do_rpc('hg', sha_not_exists, sha3)
    with pytest.raises(grpc.RpcError) as exc_info_git:
        do_rpc('git', sha_not_exists, git_shas[sha3])
    assert exc_info_hg.value.code() == exc_info_git.value.code()
    assert exc_info_hg.value.details() == exc_info_git.value.details()


def test_compare_diff_stats(gitaly_comparison):
    fixture = gitaly_comparison
    gitaly_repo = fixture.gitaly_repo
    git_repo = fixture.git_repo
    wrapper = fixture.hg_repo_wrapper
    gl_branch = b'branch/default'

    ctx0 = wrapper.commit_file('bar',
                               content="first_line\n"
                                       "second_line\n"
                                       "third_line\n",
                               message="Add bar")
    git_sha0 = git_repo.branches()[gl_branch]['sha']
    wrapper.commit_file('bar',
                        content="first_line\n"
                                "second_line\n"
                                "third_line_updated\n",
                        message="Changes bar")
    wrapper.command(b'mv', wrapper.repo.root + b'/bar',
                    wrapper.repo.root + b'/zar')
    wrapper.commit([b'zar', b'bar'], message=b"Rename bar to zar")
    ctx3 = wrapper.commit_file('zoo', content="I am in zoo\n",
                               message="Added zoo")
    git_sha3 = git_repo.branches()[gl_branch]['sha']
    # Repo structure:
    #
    # @  3 Added zoo
    # |
    # o  2 Rename bar to zar
    # |
    # o  1 Changes bar
    # |
    # o  0 Add bar
    #

    def do_rpc(vcs, left_cid, right_cid, **opts):
        request = DiffStatsRequest(
                    repository=gitaly_repo,
                    left_commit_id=left_cid,
                    right_commit_id=right_cid,
                    **opts,
                )
        diff_stubs = dict(git=DiffServiceStub(fixture.gitaly_channel),
                          hg=DiffServiceStub(fixture.hgitaly_channel))
        response = diff_stubs[vcs].DiffStats(request)
        return [resp.stats for resp in response]

    # case 1: actual test
    hg_resp = do_rpc('hg', left_cid=ctx0.hex(),
                     right_cid=ctx3.hex())
    git_resp = do_rpc('git', left_cid=git_sha0,
                      right_cid=git_sha3)
    assert hg_resp == git_resp

    # case 2: when commit_id does not correspond to a commit
    sha_not_exists = b'deadnode' * 5
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        do_rpc('hg', sha_not_exists, ctx3.hex())
    with pytest.raises(grpc.RpcError) as exc_info_git:
        do_rpc('git', sha_not_exists, git_sha3)
    assert exc_info_hg.value.code() == exc_info_git.value.code()
    assert exc_info_hg.value.details() == exc_info_git.value.details()


def test_compare_find_changed_paths(gitaly_comparison):
    fixture = gitaly_comparison
    wrapper_repo = fixture.gitaly_repo
    git_repo = fixture.git_repo
    wrapper = fixture.hg_repo_wrapper

    gl_branch = b'branch/default'
    (wrapper.path / 'sub').mkdir()
    (wrapper.path / 'sub/foo').write_text('foo content')
    (wrapper.path / 'bar').write_text('bar content')
    ctx0 = wrapper.commit(rel_paths=['sub/foo', 'bar'],
                          add_remove=True)
    git_sha0 = git_repo.branches()[gl_branch]['sha']

    (wrapper.path / 'zoo').write_text('zoo content')
    (wrapper.path / 'zoo').chmod(0o755)
    (wrapper.path / 'sub/foo').write_text('foo content modified')
    (wrapper.path / 'bar').unlink()
    wrapper.command(b'rm', wrapper.repo.root + b'/bar')
    ctx1 = wrapper.commit(rel_paths=['sub/foo', 'bar', 'zoo'],
                          add_remove=True)
    git_sha1 = git_repo.branches()[gl_branch]['sha']

    diff_stubs = dict(
        git=DiffServiceStub(fixture.gitaly_channel),
        hg=DiffServiceStub(fixture.hgitaly_channel)
    )

    oid_to_git = oid_mapping(fixture,
                             ((ctx.hex(), path, 'blob')
                              for ctx, path in (
                                      (ctx0, b'sub/foo'),
                                      (ctx1, b'sub/foo'),
                                      (ctx0, b'bar'),
                                      (ctx1, b'zoo'),
                              )))
    oid_to_git[NULL_BLOB_OID] = NULL_BLOB_OID

    def do_rpc_inner(vcs, **req_kwargs):
        """Transitional to accomodate deprecated and new style"""

        request = FindChangedPathsRequest(repository=wrapper_repo,
                                          **req_kwargs)
        response = diff_stubs[vcs].FindChangedPaths(request)
        final = []
        for resp in response:
            paths = sorted(resp.paths, key=lambda o: o.path)
            final.append(paths)
            if vcs == 'hg':
                for path in paths:
                    if path.old_blob_id:
                        path.old_blob_id = oid_to_git[path.old_blob_id]
                    if path.new_blob_id:
                        path.new_blob_id = oid_to_git[path.new_blob_id]
        return final

    def do_rpc_depr(vcs, commits):
        return do_rpc_inner(vcs, commits=commits)

    CommitRequest = FindChangedPathsRequest.Request.CommitRequest

    def do_rpc_commits(vcs, commits, compare_to=()):
        """Using the new (as of Gitaly 15.2 CommitRequest message.

        for more cases, use do_rpc_inner directly
        :param compare_to: if given, will be used in all requests.
        """
        return do_rpc_inner(
            vcs,
            requests=[
                FindChangedPathsRequest.Request(
                    commit_request=CommitRequest(
                        commit_revision=c,
                        parent_commit_revisions=compare_to))
                for c in commits]
        )

    TreeRequest = FindChangedPathsRequest.Request.TreeRequest

    def git_tree(revision, path):
        stub = CommitServiceStub(fixture.gitaly_channel)
        resps = list(stub.TreeEntry(
            TreeEntryRequest(repository=wrapper_repo,
                             revision=revision,
                             path=path)
        ))
        assert len(resps) == 1
        assert resps[0].type == TreeEntryResponse.ObjectType.TREE
        return resps[0].oid

    def hg_tree(revision, path):
        repo = wrapper.repo
        return tree_oid(repo,
                        gitlab_revision_hash(repo, revision).decode('ascii'),
                        path)

    tree_finders = dict(hg=hg_tree, git=git_tree)

    def do_rpc_tree(vcs, left_commit, right_commit, path):
        # for now we support only the case where all trees are the same subdir
        # in different commits.

        tree_finder = tree_finders[vcs]
        return do_rpc_inner(
            vcs,
            requests=[
                FindChangedPathsRequest.Request(
                    tree_request=TreeRequest(
                        left_tree_revision=tree_finder(left_commit, path),
                        right_tree_revision=tree_finder(right_commit, path))
                )
            ]
        )

    assert (do_rpc_tree('git', git_sha0, git_sha1, b'sub')
            ==
            do_rpc_tree('hg', ctx0.hex(), ctx1.hex(), b'sub')
            )

    # case 1: actual test
    assert do_rpc_depr('git', [git_sha1]) == do_rpc_depr('hg', [ctx1.hex()])
    assert (do_rpc_commits('git', [git_sha1])
            ==
            do_rpc_commits('hg', [ctx1.hex()])
            )

    # Note: khanchi97: As per the command used by gitaly for FindChangedPaths
    # cmd: `git diff-tree --stdin -z -m -r --name-status --no-renames`
    #      `--no-commit-id --diff-filter=AMDTC`
    # -> git doesn't return anything for the first commit, although I assume
    # it should, so not testing for ctx0 for now.

    # case 2: when commit_id does not correspond to a commit
    sha_not_exists = b'deadnode' * 5

    with pytest.raises(grpc.RpcError) as exc_info_hg:
        do_rpc_depr('hg', [ctx0.hex(), sha_not_exists])
    with pytest.raises(grpc.RpcError) as exc_info_git:
        do_rpc_depr('git', [git_sha0, sha_not_exists])
    assert exc_info_hg.value.code() == exc_info_git.value.code()
    assert exc_info_hg.value.details() == exc_info_git.value.details()

    with pytest.raises(grpc.RpcError) as exc_info_hg:
        do_rpc_commits('hg', [ctx0.hex(), sha_not_exists])
    with pytest.raises(grpc.RpcError) as exc_info_git:
        do_rpc_commits('git', [git_sha0, sha_not_exists])
    assert exc_info_hg.value.code() == exc_info_git.value.code()
    assert exc_info_hg.value.details() == exc_info_git.value.details()

    # case 3: tree requests
    assert (do_rpc_tree('git', git_sha0, git_sha1, b'sub')
            ==
            do_rpc_tree('hg', ctx0.hex(), ctx1.hex(), b'sub')
            )

    # can't compare with copies because Gitaly actually fails to
    # detect them (it does not pass the `-C` flag to `git diff-tree`,
    # at least as of gitaly@2b069d853)
    # not sure if that counts as a bug on their side or is really
    # intended, even though the copied case is part of protocol.

    # case 4: several requests, checking our analysis that results are
    # not merged.
    # Let's begin with the extreme case, twice the same commit:
    assert (do_rpc_commits('git', [git_sha1, git_sha1])
            ==
            do_rpc_commits('hg', [ctx1.hex(), ctx1.hex()])
            )
    # Less disturbing, let's make a new commit, modifying `foo` once again
    (wrapper.path / 'sub/foo').write_text('foo content re modified')
    hg_sha2 = wrapper.commit(rel_paths=['sub']).hex()
    git_sha2 = git_repo.branches()[gl_branch]['sha']
    fixture.invalidate()

    oid_to_git.update(oid_mapping(fixture,
                                  [(hg_sha2, b'sub/foo', 'blob')]))
    assert (do_rpc_commits('git', [git_sha1, git_sha2])
            ==
            do_rpc_commits('hg', [ctx1.hex(), hg_sha2])
            )

    # case 4: merge
    # In case of several parents, `git diff-tree -m`, on which the Gitaly
    # implementation is based would display the diffs against each parent,
    # one after the other.

    other_ctx = wrapper.commit_file('new', branch='other', parent=ctx0,
                                    content="new file in branch")
    wrapper.update(2)
    wrapper.command('merge', other_ctx.hex())
    # the foo file will appeared to be modified when compared with each parent
    (wrapper.path / 'sub/foo').write_text('something else entirely')
    # TODO implement comiit with "always matcher" in mercurial-testhelpers
    # (merge commits don't allow specifying paths, even if we list all)
    wrapper.command('commit', message=b'the merge')
    hg_sha3 = wrapper.repo[b'tip'].hex()
    git_sha3 = git_repo.branches()[gl_branch]['sha']
    fixture.invalidate()

    oid_to_git.update(
        oid_mapping(fixture,
                    ((sha, path, 'blob') for sha, path in (
                        (hg_sha3, b'new'),
                        (hg_sha3, b'sub/foo'),
                        (hg_sha3, b'zoo'),
                        (other_ctx.hex(), b'bar'),
                        (other_ctx.hex(), b'new'),
                        (other_ctx.hex(), b'sub/foo'),
                        ))))
    assert (do_rpc_commits('git', [git_sha3])
            ==
            do_rpc_commits('hg', [hg_sha3])
            )

    # case 5: passing "parents" explicitely
    hg_sha4 = wrapper.commit_file("toto").hex()
    git_sha4 = git_repo.branches()[gl_branch]['sha']
    fixture.invalidate()

    oid_to_git.update(
        oid_mapping(fixture,
                    ((sha, path, 'blob') for sha, path in (
                        (hg_sha4, b'new'),
                        (hg_sha4, b'sub/foo'),
                        (hg_sha4, b'toto'),
                        (hg_sha4, b'zoo'),
                        ))))
    assert (do_rpc_commits('git', [git_sha4], compare_to=[git_sha0])
            ==
            do_rpc_commits('hg', [hg_sha4], compare_to=[ctx0.hex()])
            )


def test_compare_get_patch_id(gitaly_comparison):
    fixture = gitaly_comparison
    wrapper = fixture.hg_repo_wrapper
    rpc_helper = fixture.rpc_helper(
        stub_cls=DiffServiceStub,
        method_name='GetPatchID',
        request_cls=GetPatchIDRequest,
        request_sha_attrs=['old_revision', 'new_revision'],
    )
    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    hex0 = wrapper.commit_file('regular', content='foo\n').hex()
    hex1 = wrapper.commit_file('regular', content='bar\n', topic='bar').hex()

    gl_branch = b'branch/default'
    gl_topic = b'topic/default/bar'

    # regular operation with symbolic or sha revisions
    assert_compare(old_revision=gl_branch, new_revision=gl_topic)
    assert_compare(old_revision=hex0, new_revision=gl_topic)
    assert_compare(old_revision=gl_branch, new_revision=hex1)

    # errors on unknown revisions (for now INTERNAL with Gitaly, highly
    # probable that it will become something else and that details really
    # do not matter)
    assert_compare_errors(old_revision=gl_branch, new_revision=b'unknown',
                          same_details=False)
    assert_compare_errors(old_revision=b'unknown', new_revision=gl_branch,
                          same_details=False)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import blob_pb2 as blob__pb2


class BlobServiceStub(object):
    """BlobService is a service which provides RPCs to retrieve Git blobs from a
    specific repository.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetBlob = channel.unary_stream(
                '/gitaly.BlobService/GetBlob',
                request_serializer=blob__pb2.GetBlobRequest.SerializeToString,
                response_deserializer=blob__pb2.GetBlobResponse.FromString,
                )
        self.GetBlobs = channel.unary_stream(
                '/gitaly.BlobService/GetBlobs',
                request_serializer=blob__pb2.GetBlobsRequest.SerializeToString,
                response_deserializer=blob__pb2.GetBlobsResponse.FromString,
                )
        self.ListBlobs = channel.unary_stream(
                '/gitaly.BlobService/ListBlobs',
                request_serializer=blob__pb2.ListBlobsRequest.SerializeToString,
                response_deserializer=blob__pb2.ListBlobsResponse.FromString,
                )
        self.ListAllBlobs = channel.unary_stream(
                '/gitaly.BlobService/ListAllBlobs',
                request_serializer=blob__pb2.ListAllBlobsRequest.SerializeToString,
                response_deserializer=blob__pb2.ListAllBlobsResponse.FromString,
                )
        self.GetLFSPointers = channel.unary_stream(
                '/gitaly.BlobService/GetLFSPointers',
                request_serializer=blob__pb2.GetLFSPointersRequest.SerializeToString,
                response_deserializer=blob__pb2.GetLFSPointersResponse.FromString,
                )
        self.ListLFSPointers = channel.unary_stream(
                '/gitaly.BlobService/ListLFSPointers',
                request_serializer=blob__pb2.ListLFSPointersRequest.SerializeToString,
                response_deserializer=blob__pb2.ListLFSPointersResponse.FromString,
                )
        self.ListAllLFSPointers = channel.unary_stream(
                '/gitaly.BlobService/ListAllLFSPointers',
                request_serializer=blob__pb2.ListAllLFSPointersRequest.SerializeToString,
                response_deserializer=blob__pb2.ListAllLFSPointersResponse.FromString,
                )


class BlobServiceServicer(object):
    """BlobService is a service which provides RPCs to retrieve Git blobs from a
    specific repository.
    """

    def GetBlob(self, request, context):
        """GetBlob returns the contents of a blob object referenced by its object
        ID. We use a stream to return a chunked arbitrarily large binary
        response
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBlobs(self, request, context):
        """GetBlobs returns blobs identified via a revision and path.

        Note that the behaviour of this RPC is quite weird: it does not only return blobs, but will also return submodules
        as commits and trees. It's use is thus discouraged in favor of ListBlobs, which behaves more sanely.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListBlobs(self, request, context):
        """ListBlobs will list all blobs reachable from a given set of revisions by
        doing a graph walk. It is not valid to pass revisions which do not resolve
        to an existing object.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListAllBlobs(self, request, context):
        """ListAllBlobs retrieves all blobs pointers in the repository, including
        those not reachable by any reference.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetLFSPointers(self, request, context):
        """GetLFSPointers retrieves LFS pointers from a given set of object IDs.
        This RPC filters all requested objects and only returns those which refer
        to a valid LFS pointer.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListLFSPointers(self, request, context):
        """ListLFSPointers retrieves LFS pointers reachable from a given set of
        revisions by doing a graph walk. This includes both normal revisions like
        an object ID or branch, but also the pseudo-revisions "--all" and "--not"
        as documented in git-rev-parse(1). Revisions which don't directly or
        transitively reference any LFS pointers are ignored. It is not valid to
        pass revisions which do not resolve to an existing object.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListAllLFSPointers(self, request, context):
        """ListAllLFSPointers retrieves all LFS pointers in the repository, including
        those not reachable by any reference.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_BlobServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetBlob': grpc.unary_stream_rpc_method_handler(
                    servicer.GetBlob,
                    request_deserializer=blob__pb2.GetBlobRequest.FromString,
                    response_serializer=blob__pb2.GetBlobResponse.SerializeToString,
            ),
            'GetBlobs': grpc.unary_stream_rpc_method_handler(
                    servicer.GetBlobs,
                    request_deserializer=blob__pb2.GetBlobsRequest.FromString,
                    response_serializer=blob__pb2.GetBlobsResponse.SerializeToString,
            ),
            'ListBlobs': grpc.unary_stream_rpc_method_handler(
                    servicer.ListBlobs,
                    request_deserializer=blob__pb2.ListBlobsRequest.FromString,
                    response_serializer=blob__pb2.ListBlobsResponse.SerializeToString,
            ),
            'ListAllBlobs': grpc.unary_stream_rpc_method_handler(
                    servicer.ListAllBlobs,
                    request_deserializer=blob__pb2.ListAllBlobsRequest.FromString,
                    response_serializer=blob__pb2.ListAllBlobsResponse.SerializeToString,
            ),
            'GetLFSPointers': grpc.unary_stream_rpc_method_handler(
                    servicer.GetLFSPointers,
                    request_deserializer=blob__pb2.GetLFSPointersRequest.FromString,
                    response_serializer=blob__pb2.GetLFSPointersResponse.SerializeToString,
            ),
            'ListLFSPointers': grpc.unary_stream_rpc_method_handler(
                    servicer.ListLFSPointers,
                    request_deserializer=blob__pb2.ListLFSPointersRequest.FromString,
                    response_serializer=blob__pb2.ListLFSPointersResponse.SerializeToString,
            ),
            'ListAllLFSPointers': grpc.unary_stream_rpc_method_handler(
                    servicer.ListAllLFSPointers,
                    request_deserializer=blob__pb2.ListAllLFSPointersRequest.FromString,
                    response_serializer=blob__pb2.ListAllLFSPointersResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gitaly.BlobService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class BlobService(object):
    """BlobService is a service which provides RPCs to retrieve Git blobs from a
    specific repository.
    """

    @staticmethod
    def GetBlob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.BlobService/GetBlob',
            blob__pb2.GetBlobRequest.SerializeToString,
            blob__pb2.GetBlobResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetBlobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.BlobService/GetBlobs',
            blob__pb2.GetBlobsRequest.SerializeToString,
            blob__pb2.GetBlobsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListBlobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.BlobService/ListBlobs',
            blob__pb2.ListBlobsRequest.SerializeToString,
            blob__pb2.ListBlobsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListAllBlobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.BlobService/ListAllBlobs',
            blob__pb2.ListAllBlobsRequest.SerializeToString,
            blob__pb2.ListAllBlobsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetLFSPointers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.BlobService/GetLFSPointers',
            blob__pb2.GetLFSPointersRequest.SerializeToString,
            blob__pb2.GetLFSPointersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListLFSPointers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.BlobService/ListLFSPointers',
            blob__pb2.ListLFSPointersRequest.SerializeToString,
            blob__pb2.ListLFSPointersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListAllLFSPointers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.BlobService/ListAllLFSPointers',
            blob__pb2.ListAllLFSPointersRequest.SerializeToString,
            blob__pb2.ListAllLFSPointersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import server_pb2 as server__pb2


class ServerServiceStub(object):
    """ServerService is a service that provides information about a Gitaly server.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ServerInfo = channel.unary_unary(
                '/gitaly.ServerService/ServerInfo',
                request_serializer=server__pb2.ServerInfoRequest.SerializeToString,
                response_deserializer=server__pb2.ServerInfoResponse.FromString,
                )
        self.DiskStatistics = channel.unary_unary(
                '/gitaly.ServerService/DiskStatistics',
                request_serializer=server__pb2.DiskStatisticsRequest.SerializeToString,
                response_deserializer=server__pb2.DiskStatisticsResponse.FromString,
                )
        self.ClockSynced = channel.unary_unary(
                '/gitaly.ServerService/ClockSynced',
                request_serializer=server__pb2.ClockSyncedRequest.SerializeToString,
                response_deserializer=server__pb2.ClockSyncedResponse.FromString,
                )
        self.ReadinessCheck = channel.unary_unary(
                '/gitaly.ServerService/ReadinessCheck',
                request_serializer=server__pb2.ReadinessCheckRequest.SerializeToString,
                response_deserializer=server__pb2.ReadinessCheckResponse.FromString,
                )


class ServerServiceServicer(object):
    """ServerService is a service that provides information about a Gitaly server.
    """

    def ServerInfo(self, request, context):
        """ServerInfo ...
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DiskStatistics(self, request, context):
        """DiskStatistics ...
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ClockSynced(self, request, context):
        """ClockSynced checks if machine clock is synced
        (the offset is less that the one passed in the request).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReadinessCheck(self, request, context):
        """ReadinessCheck runs the set of the checks to make sure service is in operational state.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ServerServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ServerInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInfo,
                    request_deserializer=server__pb2.ServerInfoRequest.FromString,
                    response_serializer=server__pb2.ServerInfoResponse.SerializeToString,
            ),
            'DiskStatistics': grpc.unary_unary_rpc_method_handler(
                    servicer.DiskStatistics,
                    request_deserializer=server__pb2.DiskStatisticsRequest.FromString,
                    response_serializer=server__pb2.DiskStatisticsResponse.SerializeToString,
            ),
            'ClockSynced': grpc.unary_unary_rpc_method_handler(
                    servicer.ClockSynced,
                    request_deserializer=server__pb2.ClockSyncedRequest.FromString,
                    response_serializer=server__pb2.ClockSyncedResponse.SerializeToString,
            ),
            'ReadinessCheck': grpc.unary_unary_rpc_method_handler(
                    servicer.ReadinessCheck,
                    request_deserializer=server__pb2.ReadinessCheckRequest.FromString,
                    response_serializer=server__pb2.ReadinessCheckResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gitaly.ServerService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ServerService(object):
    """ServerService is a service that provides information about a Gitaly server.
    """

    @staticmethod
    def ServerInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.ServerService/ServerInfo',
            server__pb2.ServerInfoRequest.SerializeToString,
            server__pb2.ServerInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DiskStatistics(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.ServerService/DiskStatistics',
            server__pb2.DiskStatisticsRequest.SerializeToString,
            server__pb2.DiskStatisticsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ClockSynced(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.ServerService/ClockSynced',
            server__pb2.ClockSyncedRequest.SerializeToString,
            server__pb2.ClockSyncedResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ReadinessCheck(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.ServerService/ReadinessCheck',
            server__pb2.ReadinessCheckRequest.SerializeToString,
            server__pb2.ReadinessCheckResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
NO_LIMIT = 2147483647


def extract_limit(request, field_name='pagination_params') -> int:
    """Extract limit with our conventions from given pagination_params.
    For context, gRPC does not have any notion of ``None``: absence
    of value is conflated with default value (which is 0 for integers).

    For specification, see comment in `shared.proto`.
    """
    if not request.HasField(field_name):
        return NO_LIMIT

    limit = getattr(request, field_name).limit
    # 0 really means to ask for empty response (it could have meant no
    # limit, the usual dilemma with integers in a language without None,
    # such as ProtoBuf)
    if limit < 0:
        limit = NO_LIMIT
    return limit



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

# TODO bring full coverage directly from here, because there might
# be points in time where it won't be used at all by gRPC methods
import grpc
from grpc.beta import _metadata


class FakeServicerContext(grpc.ServicerContext):
    """A base class for fake servicer contexts in tests.

    This base class allow testing harness to implement only a few methods,
    while still going through the ABC checking that all abstract methods are
    implemented.

    This can detect API changes in the base class. For an example, assume
    that main code relies on the `invocation_metadata` method,
    and corresponding unit  test provides a mock implementation. If the method
    is renamed in :class:`ServicerContext` as `incoming_metadata`, the
    mock implementation suddenly will not implement all abstract methods
    and the test run will fail, alerting us of the needed adaptation.

    This is heavy-handed, but still a bit better than not doing any checking.
    Of course most of HGItaly is covered by integration tests such as
    in `hgitaly.service.tests`, but some might temporarily not be
    (concrete case being the feature flags system at a time where there
    is no relevant Gitaly feature flag for HGitaly), and hence requires
    full coverage in *unit* tests.
    """

    def fake_impl(self):
        """To make ABC consider an abstract method to be implemented."""
        raise NotImplementedError()  # pragma no cover

    abort = fake_impl
    abort_with_status = fake_impl
    add_callback = fake_impl
    auth_context = fake_impl
    cancel = fake_impl
    invocation_metadata = fake_impl
    is_active = fake_impl
    peer = fake_impl
    peer_identities = fake_impl
    peer_identity_key = fake_impl
    send_initial_metadata = fake_impl
    set_code = fake_impl
    set_details = fake_impl
    set_trailing_metadata = fake_impl
    time_remaining = fake_impl


class FakeContextAbort(RuntimeError):
    pass


class FakeContextAborter(FakeServicerContext):
    """A fake context that is good enough to test various `abort()` cases.
    """

    _trailing_metadata = _code = _details = None

    def set_trailing_metadata(self, metadata):
        self._trailing_metadata = metadata

    def trailing_metadata(self):
        return self._trailing_metadata

    def code(self):
        return self._code

    def details(self):
        return self._details

    def abort(self, code, details):
        self._code = code
        self._details = details
        raise FakeContextAbort()


def metadatum(key, val):
    return _metadata._Metadatum(key, val)



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import gc
import grpc
import os
from pathlib import Path
import pytest
from mercurial import (
    hg,
    pycompat,
)

from heptapod import obsutil

from mercurial_testhelpers import (
    as_bytes,
)

from heptapod.testhelpers import (
    LocalRepoWrapper,
)

from ..testing.context import (
    FakeServicerContext,
    metadatum,
)
from ..servicer import (
    HG_GIT_MIRRORING_MD_KEY,
    NATIVE_PROJECT_MD_KEY,
    ACCEPT_MR_IID_KEY,
    SKIP_HOOKS_MD_KEY,
    PY_HEPTAPOD_SKIP_HOOKS,
    HGitalyServicer,
)
from ..stub.shared_pb2 import (
    Repository,
    User,
)
from ..stub.repository_pb2 import HasLocalBranchesRequest
from ..stub.repository_pb2_grpc import RepositoryServiceStub


class AbortContext(Exception):
    """Raised by FakeContext.abort.

    gRPC's context.abort raises `Exception()` (sic), which is
    inconvenient for testing.
    """


class FakeContext(FakeServicerContext):

    _invocation_metadata = ()

    def abort(self, code, message):
        self.code = code
        self.message = message
        raise AbortContext()

    def set_code(self, code):
        self.code = code

    def set_details(self, details):
        self.details = details

    def invocation_metadata(self):
        return self._invocation_metadata

    def set_invocation_md(self, md):
        self._invocation_metadata += (md, )


def test_load_repo(tmpdir):
    storage_root = tmpdir.join('repos')
    storage_root_bytes = pycompat.sysbytes(str(storage_root))
    servicer = HGitalyServicer(dict(storname=storage_root_bytes))
    # context is used for error raising only
    context = FakeContext()

    wrapper = LocalRepoWrapper.init(storage_root.join('awesome-proj.hg'))
    loaded = servicer.load_repo(Repository(storage_name='storname',
                                           relative_path='awesome-proj.hg'),
                                context)
    assert loaded.root == wrapper.repo.root

    # In practice, requests from the Rails app will assume the relevant
    # path to end in `.git`, we need to ignore that.
    loaded = servicer.load_repo(Repository(storage_name='storname',
                                           relative_path='awesome-proj.git'),
                                context)
    assert loaded.root == wrapper.repo.root

    with pytest.raises(KeyError) as exc_info:
        servicer.load_repo_inner(Repository(storage_name='dream',
                                            relative_path='dream-proj.hg'),
                                 context)
    assert exc_info.value.args == ('storage', 'dream')

    with pytest.raises(AbortContext):
        servicer.load_repo(Repository(storage_name='dream',
                                      relative_path='dream-proj.hg'),
                           context)
    assert context.code == grpc.StatusCode.NOT_FOUND


def test_load_repo_for_mutation(tmpdir):
    storage_root = tmpdir.join('repos')
    storage_root_bytes = pycompat.sysbytes(str(storage_root))
    servicer = HGitalyServicer(dict(storname=storage_root_bytes))
    LocalRepoWrapper.init(storage_root.join('awesome-proj.hg'))
    grpc_repo = Repository(storage_name='storname',
                           relative_path='awesome-proj.hg')
    user = User(gl_id='whatever', name=b"Georges", email=b'gr@heptapod.test')

    context = FakeContext()
    loaded = servicer.load_repo(grpc_repo, context, for_mutation_by=user)
    assert loaded.ui.environ[b'HEPTAPOD_USERINFO_GL_ID'] == b'whatever'
    assert loaded.ui.configbool(b'heptapod', b'native') is False
    assert loaded.ui.configbool(b'heptapod', b'no-git') is False

    context.set_invocation_md(metadatum(NATIVE_PROJECT_MD_KEY, 'True'))
    loaded = servicer.load_repo(grpc_repo, context, for_mutation_by=user)
    assert loaded.ui.configbool(b'heptapod', b'native') is True
    assert loaded.ui.configbool(b'heptapod', b'no-git') is True

    context.set_invocation_md(metadatum(HG_GIT_MIRRORING_MD_KEY, 'True'))
    loaded = servicer.load_repo(grpc_repo, context, for_mutation_by=user)
    assert loaded.ui.configbool(b'heptapod', b'native') is True
    assert loaded.ui.configbool(b'heptapod', b'no-git') is False

    context.set_invocation_md(metadatum(ACCEPT_MR_IID_KEY, '609'))
    loaded = servicer.load_repo(grpc_repo, context, for_mutation_by=user)
    assert loaded.ui.environ[b'HEPTAPOD_ACCEPT_MR_IID'] == b'609'

    context.set_invocation_md(metadatum(SKIP_HOOKS_MD_KEY, 'True'))
    loaded = servicer.load_repo(grpc_repo, context, for_mutation_by=user)
    assert loaded.ui.environ[PY_HEPTAPOD_SKIP_HOOKS] == b'yes'

    with pytest.raises(AbortContext):
        servicer.load_repo(grpc_repo, context, for_mutation_by=User())
    assert context.code == grpc.StatusCode.INVALID_ARGUMENT


def test_load_repo_gc(tmpdir):
    storage_root = tmpdir.join('repos')
    storage_root_bytes = pycompat.sysbytes(str(storage_root))
    servicer = HGitalyServicer(dict(storname=storage_root_bytes))
    # context is used for error raising only
    context = FakeContext()

    src_path = storage_root.join('awesome-proj.hg')
    dst_path = storage_root.join('shared.hg')
    LocalRepoWrapper.init(src_path)
    LocalRepoWrapper.share_from_path(src_path, dst_path)

    from mercurial.repoview import _filteredrepotypes as frt
    # we're checking that stuff doesn't accumulate in this frt
    # since it is a cache, we can start with a clean slate:
    frt.clear()

    loaded = servicer.load_repo(Repository(storage_name='storname',
                                           relative_path='shared.hg'),
                                context)
    # not what we really need, but will help demonstrate that proper removal
    # from frt actually happens.
    assert len(frt) > 0

    # let's even make a loop of references:
    hg.sharedreposource(loaded).dst_repo = loaded

    del loaded
    gc.collect()

    # this may turn out to be optimistic
    assert len(frt) == 0


def test_errors_propagation(grpc_channel, server_repos_root):
    # Taking a random RPC to check that the client receives the
    # proper error response
    repo_stub = RepositoryServiceStub(grpc_channel)

    with pytest.raises(grpc.RpcError) as exc_info:
        repo_stub.HasLocalBranches(HasLocalBranchesRequest(
            repository=Repository(storage_name='dream', relative_path='')))
    exc = exc_info.value

    assert exc.code() == grpc.StatusCode.INVALID_ARGUMENT
    assert 'dream' in exc.details()

    with pytest.raises(grpc.RpcError) as exc_info:
        repo_stub.HasLocalBranches(HasLocalBranchesRequest(
            repository=Repository(storage_name='default',
                                  relative_path='not_here')))
    exc = exc_info.value

    assert exc.code() == grpc.StatusCode.NOT_FOUND
    assert 'not_here' in exc.details()


def test_temp_dir(tmpdir):
    # context is used for error raising only
    context = FakeContext()
    servicer = HGitalyServicer(dict(default=as_bytes(tmpdir)))

    path = servicer.temp_dir('default', context, ensure=False)
    path = Path(os.fsdecode(path))
    assert path.is_relative_to(tmpdir)
    assert not path.exists()

    assert servicer.temp_dir('default', context, ensure=True) == bytes(path)
    assert path.exists()

    with pytest.raises(KeyError) as exc_info:
        servicer.temp_dir_inner('unknown', context)
    assert exc_info.value.args == ('storage', 'unknown')

    with pytest.raises(AbortContext) as exc_info:
        servicer.temp_dir('unknown', context)
    assert context.code == grpc.StatusCode.NOT_FOUND

    # this is how it looks with a missing repository argument in the
    # gRPC request, and temp_dir(request.repository.storage_name)
    with pytest.raises(AbortContext) as exc_info:
        servicer.temp_dir('', context)
    assert context.code == grpc.StatusCode.INVALID_ARGUMENT


def test_temp_dir_failed_creation(tmpdir):
    # context is used for error raising only
    context = FakeContext()
    broken = tmpdir / 'broken'
    broken.mksymlinkto('broken')  # itself

    servicer = HGitalyServicer(dict(broken=as_bytes(broken)))

    with pytest.raises(AbortContext):
        servicer.temp_dir('broken', context, ensure=True)
    assert context.code == grpc.StatusCode.INTERNAL


def test_working_dir(tmpdir):
    context = FakeContext()
    servicer = HGitalyServicer(dict(default=as_bytes(tmpdir)))

    storage_root = tmpdir.join('repos')
    storage_root_bytes = pycompat.sysbytes(str(storage_root))
    servicer = HGitalyServicer(dict(storname=storage_root_bytes))
    # context is used for error raising only
    context = FakeContext()

    wrapper = LocalRepoWrapper.init(storage_root.join('awesome-proj.hg'))
    gl_repo = Repository(storage_name='storname',
                         relative_path='awesome-proj.hg')
    repo = servicer.load_repo(gl_repo, context)

    with servicer.working_dir(gl_repo, repo, context) as wd:
        wd_path = wd.path
        wd_wrapper = LocalRepoWrapper.load(wd_path)
        sha = wd_wrapper.commit_file('foo').hex()

    wrapper.reload()
    ctx = wrapper.repo[sha]
    assert ctx.hex() == sha

    with servicer.working_dir(gl_repo, repo, context, changeset=ctx) as wd:
        assert wd.path == wd_path
        wd_wrapper.reload()
        sha2 = wd_wrapper.commit_file('nar').hex()

    wrapper.reload()
    ctx2 = wrapper.repo[sha2]
    assert ctx2.hex() == sha2


def test_repo_command(tmpdir):
    storage_root = tmpdir.join('repos')
    storage_root_bytes = pycompat.sysbytes(str(storage_root))
    servicer = HGitalyServicer(dict(storname=storage_root_bytes))
    wrapper = LocalRepoWrapper.init(storage_root.join('awesome-proj.hg'))
    ctx = wrapper.commit_file('foo')

    repo = wrapper.repo
    context = FakeContext()

    servicer.repo_command(repo, context, 'amend', message=b"amended")
    ctx = wrapper.repo.unfiltered()[ctx.rev()]
    amended = obsutil.latest_unique_successor(ctx)
    assert amended.description() == b'amended'



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import gc
from grpc import StatusCode
from io import BytesIO
import itertools
import logging
import os
from collections import deque
import re
import shutil
import tempfile
import time

from mercurial import (
    archival,
    pycompat,
    scmutil,
    node,
)
from mercurial.commands import (
    bundle,
)

from heptapod.gitlab.branch import gitlab_branch_from_ref
from hgext3rd.heptapod import (
    backup_additional,
    restore_additional,
)
from hgext3rd.heptapod.branch import set_default_gitlab_branch
from hgext3rd.heptapod.special_ref import write_gitlab_special_ref
from hgext3rd.heptapod.keep_around import (
    create_keep_around,
    parse_keep_around_ref,
)

from .. import (
    manifest,
    message,
)
from ..branch import (
    iter_gitlab_branches,
)
from ..errors import (
    not_implemented,
)
from ..gitlab_ref import (
    parse_special_ref,
    ensure_special_refs,
    gitlab_special_ref_target,
)
from ..license_detector import detect_license
from ..logging import LoggerAdapter
from ..path import (
    InvalidPath,
    validate_relative_path,
)
from ..repository import (
    repo_size,
    unbundle,
)
from ..revision import (
    ALL_CHANGESETS,
    CHANGESET_HASH_BYTES_REGEXP,
    ZERO_SHA,
    RevisionNotFound,
    gitlab_revision_changeset,
    gitlab_revision_hash,
    resolve_revspecs_positive_negative,
)
from ..servicer import HGitalyServicer
from ..stream import (
    WRITE_BUFFER_SIZE,
    streaming_request_tempfile_extract,
)
from ..util import (
    chunked,
)
from ..workdir import (
    WorkingDirectoryError,
    remove_all_workdirs_bare,
)
from ..stub.repository_pb2 import (
    BackupCustomHooksRequest,
    BackupCustomHooksResponse,
    CreateBundleRequest,
    CreateBundleResponse,
    CreateBundleFromRefListRequest,
    CreateBundleFromRefListResponse,
    CreateRepositoryRequest,
    CreateRepositoryResponse,
    CreateRepositoryFromBundleRequest,
    CreateRepositoryFromBundleResponse,
    FetchBundleRequest,
    FetchBundleResponse,
    FindLicenseRequest,
    FindLicenseResponse,
    FindMergeBaseRequest,
    FindMergeBaseResponse,
    GetRawChangesRequest,
    GetRawChangesResponse,
    GetCustomHooksRequest,
    GetCustomHooksResponse,
    RepositoryExistsRequest,
    RepositoryExistsResponse,
    GetArchiveRequest,
    GetArchiveResponse,
    HasLocalBranchesRequest,
    HasLocalBranchesResponse,
    ObjectFormatRequest,
    ObjectFormatResponse,
    RemoveRepositoryRequest,
    RemoveRepositoryResponse,
    RepositorySizeRequest,
    RepositorySizeResponse,
    RestoreCustomHooksRequest,
    RestoreCustomHooksResponse,
    SearchFilesByContentRequest,
    SearchFilesByContentResponse,
    SearchFilesByNameRequest,
    SearchFilesByNameResponse,
    SetCustomHooksRequest,
    SetCustomHooksResponse,
    WriteRefRequest,
    WriteRefResponse,
    ApplyGitattributesRequest,
    ApplyGitattributesResponse,
)
from ..stub.repository_pb2_grpc import RepositoryServiceServicer
from ..stub.shared_pb2 import (
    ObjectFormat,
)


base_logger = logging.getLogger(__name__)
DEFAULT_BRANCH_FILE_NAME = b'default_gitlab_branch'
ARCHIVE_FORMATS = {
    GetArchiveRequest.Format.Value('ZIP'): b'zip',
    GetArchiveRequest.Format.Value('TAR'): b'tar',
    GetArchiveRequest.Format.Value('TAR_GZ'): b'tgz',
    GetArchiveRequest.Format.Value('TAR_BZ2'): b'tbz2',
}
SEARCH_FILES_FILTER_MAX_LENGTH = 1000
"""Maximum size of regular expression in SearchFiles methods.

Value taken from Gitaly's `internal/gitaly/service/repository/search_files.go'
"""


class RepositoryServicer(RepositoryServiceServicer, HGitalyServicer):
    """RepositoryServiceService implementation.
    """

    STATUS_CODE_STORAGE_NOT_FOUND = StatusCode.INVALID_ARGUMENT

    def FindMergeBase(self,
                      request: FindMergeBaseRequest,
                      context) -> FindMergeBaseResponse:
        repo = self.load_repo(request.repository, context)
        if len(request.revisions) < 2:
            context.abort(StatusCode.INVALID_ARGUMENT,
                          'at least 2 revisions are required')

        ctxs = []
        for rev in request.revisions:
            ctx = gitlab_revision_changeset(repo, rev)
            if ctx is None:
                return FindMergeBaseResponse()
            ctxs.append(ctx)

        # Some of the changesets may be obsolete (if addressed by SHAs).
        # The GCA may be obsolete as well (meaning that one of ctxs is
        # also obsolete). TODO add test for obsolete cases
        repo = repo.unfiltered()
        gca = repo.revs(b"ancestor(%ld)", ctxs).first()
        base = repo[gca].hex() if gca is not None else ''
        return FindMergeBaseResponse(base=base)

    def RepositoryExists(self,
                         request: RepositoryExistsRequest,
                         context) -> RepositoryExistsResponse:
        try:
            self.load_repo_inner(request.repository, context)
            exists = True
        except KeyError as exc:
            if exc.args[0] == 'storage':
                context.abort(
                    StatusCode.INVALID_ARGUMENT,
                    f'storage not found: "{exc.args[1]}"'
                )
            exists = False
        except ValueError as exc:
            context.abort(StatusCode.INVALID_ARGUMENT, exc.args[0])

        return RepositoryExistsResponse(exists=exists)

    def RepositorySize(self,
                       request: RepositorySizeRequest,
                       context) -> RepositorySizeResponse:
        try:
            repo = self.load_repo_inner(request.repository, context)
        except KeyError as exc:
            kind = exc.args[0]
            if kind == 'storage':
                context.abort(
                    StatusCode.INVALID_ARGUMENT,
                    f'storage not found: "{exc.args[1]}"'
                )
            elif kind == 'repo':
                context.abort(StatusCode.NOT_FOUND, "repository not found")
            raise  # pragma no cover (currently unreachable, just in case)
        except ValueError as exc:
            context.abort(StatusCode.INVALID_ARGUMENT, exc.args[0])

        return RepositorySizeResponse(size=repo_size(repo))

    def ObjectFormat(self,
                     request: ObjectFormatRequest,
                     context) -> ObjectFormatResponse:
        # we have to return an error if the repository does not exist,
        # to match Gitaly behaviour (caller could use this call to
        # check both at once).
        self.load_repo(request.repository, context)
        return ObjectFormatResponse(
            format=ObjectFormat.OBJECT_FORMAT_UNSPECIFIED)

    def GetArchive(self,
                   request: GetArchiveRequest,
                   context) -> GetArchiveResponse:
        repo = self.load_repo(request.repository, context)
        ctx = repo[request.commit_id]

        patterns = []
        path = request.path
        if path:
            try:
                path = validate_relative_path(repo, path)
            except InvalidPath:
                context.abort(StatusCode.INVALID_ARGUMENT,
                              "Invalid path: '%s'" % path)

            patterns.append(b"path:" + path)

        match = scmutil.match(ctx, pats=patterns, opts={})

        # using an anonymous (not linked) temporary file
        # TODO OPTIM check if archive is not by any chance
        # using a tempfile already…
        with tempfile.TemporaryFile(
                mode='wb+',  # the default, but let's insist on binary here
                buffering=WRITE_BUFFER_SIZE) as tmpf:
            try:
                archival.archive(
                    repo,
                    tmpf,
                    ctx.node(),
                    ARCHIVE_FORMATS[request.format],
                    # TODO this is the default but check what it means:
                    True,  # decode
                    match,
                    request.prefix.encode(),
                    subrepos=False  # maybe later, check GitLab's standard
                )
            finally:
                gc.collect()

            tmpf.seek(0)
            while True:
                data = tmpf.read(WRITE_BUFFER_SIZE)
                if not data:
                    break
                yield GetArchiveResponse(data=data)

    def HasLocalBranches(self,
                         request: HasLocalBranchesRequest,
                         context) -> HasLocalBranchesResponse:
        repo = self.load_repo(request.repository, context)
        # the iteration should stop as soon at first branchmap entry which
        # has a non closed head (but all heads in that entry would be checked
        # to be non closed)
        return HasLocalBranchesResponse(value=any(iter_gitlab_branches(repo)))

    def WriteRef(
            self,
            request: WriteRefRequest,
            context) -> WriteRefResponse:
        """Create or update a GitLab ref.

        The reference Gitaly implementation treats two cases, ``HEAD`` being
        the only supported symbolic ref. Excerpt as of GitLab 13.9.0::

          func (s *server) writeRef(ctx context.Context,
                                    req *gitalypb.WriteRefRequest) error {
            if string(req.Ref) == "HEAD" {
              return s.updateSymbolicRef(ctx, req)
            }
            return updateRef(ctx, s.cfg, s.gitCmdFactory, req)
          }

        On the other hand, the target revision is fully resolved, even
        when setting a non-symbolic ref.
        """
        logger = LoggerAdapter(base_logger, context)
        ref, target = request.ref, request.revision
        expected_old = request.old_revision
        repo = self.load_repo(request.repository, context)

        try:
            special_ref_name = parse_special_ref(ref)
            if special_ref_name is not None:
                target_sha = gitlab_revision_changeset(repo, target)
                if expected_old:
                    try:
                        expected_old_sha = gitlab_revision_hash(repo,
                                                                expected_old)
                    except RevisionNotFound:
                        # it is used mostly with hashes. If the hash is wrong
                        # (not pointing to any topic), we don't want to prevent
                        # updating the reference to fix it!
                        expected_old_sha = expected_old
                    existing_old_cs = gitlab_special_ref_target(repo, ref)
                    if expected_old == ZERO_SHA:
                        if existing_old_cs is not None:
                            context.abort(
                                StatusCode.INTERNAL,
                                "error when running update-ref command: "
                                "reference already exists")
                    elif (existing_old_cs is not None
                          and existing_old_cs.hex() != expected_old_sha):
                        context.abort(
                            StatusCode.INTERNAL,
                            "error when running update-ref command: "
                            "reference does not point to expected object")

                ensure_special_refs(repo)
                write_gitlab_special_ref(repo, special_ref_name, target_sha)
                return WriteRefResponse()

            keep_around = parse_keep_around_ref(ref)
            if keep_around is not None:
                if (CHANGESET_HASH_BYTES_REGEXP.match(keep_around) is None
                        or target != keep_around):
                    context.abort(
                        StatusCode.INVALID_ARGUMENT,
                        "Invalid target %r for keep-around %r. Only full "
                        "changeset ids in hexadecimal form are accepted and "
                        "target must "
                        "match the ref name" % (target, ref)
                    )

                create_keep_around(repo, target)
                return WriteRefResponse()
        except Exception:
            # TODO this is a stop-gap measure to prevent repository breakage
            # until we get confident enough with `ensure_special_refs` for all
            # existing repositories. This goes against all of our principles,
            # but it's more prudent this way.
            logger.exception(
                "WriteRef failed for Repository %r on storage %r",
                request.repository.relative_path,
                request.repository.storage_name)
            return WriteRefResponse()

        if ref != b'HEAD':
            context.abort(
                StatusCode.INVALID_ARGUMENT,
                "Setting ref %r is not implemented in Mercurial (target %r) "
                "Does not make sense in the case of branches and tags, "
                "except maybe for bookmarks." % (ref, target))

        # TODO also old_revision here
        target_branch = gitlab_branch_from_ref(target)
        if target_branch is None:
            context.abort(StatusCode.INVALID_ARGUMENT,
                          "The default GitLab branch can only be set "
                          "to a branch ref, got %r" % target)

        set_default_gitlab_branch(repo, target_branch)
        return WriteRefResponse()

    def ApplyGitattributes(self, request: ApplyGitattributesRequest,
                           context) -> ApplyGitattributesResponse:
        """Method used as testing bed for the `not_implemented` helper.

        It is unlikely we ever implement this one, and if we do something
        similar, we'll probably end up defining a ApplyHgAttributes anyway.
        """
        not_implemented(context, issue=1234567)

    def CreateRepository(self, request: CreateRepositoryRequest,
                         context) -> CreateRepositoryResponse:
        default_branch = request.default_branch
        try:
            repo = self.create_and_load_repo(request.repository, context)
            if default_branch:
                set_default_gitlab_branch(repo, default_branch)
        finally:
            # response is the same in case of error or success
            return CreateRepositoryResponse()

    def GetRawChanges(self, request: GetRawChangesRequest,
                      context) -> GetRawChangesResponse:
        not_implemented(context, issue=79)  # pragma no cover

    def SearchFilesByName(self, request: SearchFilesByNameRequest,
                          context) -> SearchFilesByNameResponse:
        repo = self.load_repo(request.repository, context)
        changeset = gitlab_revision_changeset(repo, request.ref)
        if changeset is None:
            yield SearchFilesByNameResponse()
            return

        query = request.query
        if not query:
            context.abort(StatusCode.INVALID_ARGUMENT, "no query given")

        subdir = b'' if query == '.' else query.encode('utf8')

        miner = manifest.miner(changeset)

        filt = request.filter
        if not filt:
            rx = None
        else:
            if len(filt) > SEARCH_FILES_FILTER_MAX_LENGTH:
                context.abort(StatusCode.INVALID_ARGUMENT,
                              "filter exceeds maximum length")
            # TODO what to do for invalid Regexp? Especially with incoming
            # format being Golang's
            try:
                rx = re.compile(filt.encode('utf8'))
            except re.error as exc:
                context.abort(StatusCode.INVALID_ARGUMENT,
                              f"filter did not compile: {exc})")

        start = request.offset
        end = None if request.limit == 0 else start + request.limit

        for paths in chunked(
                itertools.islice(
                    miner.file_names_by_regexp(rx, subdir),
                    start, end),
                force_if_empty=True,
        ):
            yield SearchFilesByNameResponse(files=paths)

    def SearchFilesByContent(self, request: SearchFilesByContentRequest,
                             context) -> SearchFilesByContentResponse:
        """Almost straight results from `git grep`

        this part of the protocol is totally undocumented, but here is what
        Gitaly does:

        - each match is sent with two lines of context before and after
        - in case of overlapping matches (including context), they are sent
          as one
        - sending a match means cutting it in chunkd of `WRITE_BUFFER_SIZE`.
        - after the full sending of each match aa  with the end_of_match`
          boolean set to true and no data is sent.

        (see also Comparison Tests for validatation of this).
        """
        repo = self.load_repo(request.repository, context)
        query = request.query
        if not query:
            context.abort(StatusCode.INVALID_ARGUMENT, "no query given")

        ref = request.ref
        changeset = gitlab_revision_changeset(repo, ref)
        if changeset is None:
            return

        # TODO filtermaxlen?
        rx = re.compile(query.encode('utf-8'), flags=re.IGNORECASE)
        # TODO chunked_response (not used by Rails app)
        match_data = BytesIO()
        miner = manifest.miner(changeset)
        for path, content, _flags in miner.iter_files_with_content(
                exclude_binary=True):
            # TODO rx options. Here's Gitaly reference (Golang), arguments
            # passed to `git grep`):
            # (with const surroundContext = 2)
            #
            #   git.Flag{Name: "--ignore-case"},
            #   git.Flag{Name: "-I"},
            #      (ignore binary files)
            #   git.Flag{Name: "--line-number"},
            #   git.Flag{Name: "--null"},
            #      (use null as delimiter between path, line number and
            #       matching line in the output, hence avoiding the need
            #       to escape)
            #   git.ValueFlag{Name: "--before-context",
            #                 Value: surroundContext},
            #   git.ValueFlag{Name: "--after-context", Value: surroundContext},
            #   git.Flag{Name: "--perl-regexp"},
            #   git.Flag{Name: "-e"},
            # Gitaly does not fill in `matches` field any more
            # (surely deprecated) and the Rails client does not read it
            # TODO cheaper iteration on a window of lines (note that
            # splitlines() takes care of `\r\n` and even `\r` EOLs.)
            for matching_lines in grep_file(rx, content):
                render_git_grep_matches(
                    match_data,
                    ref, path,
                    matching_lines)
                yield from iter_sfbc_resps(match_data)
                match_data.truncate(0)
                match_data.seek(0)

    def set_custom_hooks(self, request, context):
        def load_repo(req, context):
            return self.load_repo(req.repository, context)

        with streaming_request_tempfile_extract(
                request, context,
                first_request_handler=load_repo
        ) as (repo, tmpf):
            tmpf.flush()
            try:
                restore_additional(repo.ui, repo,
                                   tmpf.name.encode('ascii'))
            except Exception as exc:
                context.abort(StatusCode.INTERNAL,
                              "Error in tarball application: %r" % exc)

    def RestoreCustomHooks(self, request: RestoreCustomHooksRequest,
                           context) -> RestoreCustomHooksResponse:
        try:
            self.set_custom_hooks(request, context)
        finally:
            return RestoreCustomHooksResponse()

    def SetCustomHooks(self, request: SetCustomHooksRequest,
                       context) -> SetCustomHooksResponse:
        try:
            self.set_custom_hooks(request, context)
        finally:
            return SetCustomHooksResponse()

    def get_custom_hooks(self, request, context, resp_cls):
        repo = self.load_repo(request.repository, context)
        with tempfile.NamedTemporaryFile(
                mode='wb+',
                buffering=WRITE_BUFFER_SIZE) as tmpf:

            # TODO we should simply have backup_additional()
            # accept a file object rather than a path.
            save_path = pycompat.sysbytes(tmpf.name)
            backup_additional(repo.ui, repo, save_path)
            tmpf.seek(0)
            while True:
                data = tmpf.read(WRITE_BUFFER_SIZE)
                if not data:
                    break
                yield resp_cls(data=data)

    def BackupCustomHooks(self, request: BackupCustomHooksRequest,
                          context) -> BackupCustomHooksResponse:
        yield from self.get_custom_hooks(
            request, context, BackupCustomHooksResponse)

    def GetCustomHooks(self, request: GetCustomHooksRequest,
                       context) -> GetCustomHooksResponse:
        yield from self.get_custom_hooks(
            request, context, GetCustomHooksResponse)

    def RemoveRepository(self, request: RemoveRepositoryRequest,
                         context) -> RemoveRepositoryResponse:
        # The protocol comment says, as of Gitaly 14.8:
        #     RemoveRepository will move the repository to
        #     `+gitaly/tmp/<relative_path>_removed` and
        #     eventually remove it.
        # In that sentence, the "eventually" could imply that it is
        # asynchronous (as the Rails app does), but it is not in the
        # Gitaly server implementation. The renaming is done for
        # atomicity purposes.
        try:
            repo_path = self.repo_disk_path(request.repository, context)
        except KeyError as exc:
            self.handle_key_error(context, exc.args)
        except ValueError as exc:
            self.handle_value_error(context, exc.args)

        if not os.path.exists(repo_path):
            # same error message as Gitaly (probably no need to repeat
            # repo details, since request often logged client-side)
            context.abort(StatusCode.NOT_FOUND, "repository does not exist")

        trash_path = os.path.join(
            self.temp_dir(request.repository.storage_name, context),
            os.path.basename(repo_path) + b'+removed-%f' % time.time())
        # The rename being atomic, it avoids leaving a crippled repo behind
        # in case of problem in the removal.
        # TODO Gitaly also performs some kind of locking (not clear
        # if Mercurial locks would be appropriate because of the renaming)
        # and lengthy rechecks to safeguard against race conditions,
        # and finally the vote related to the multi-phase commit for praefect
        os.rename(repo_path, trash_path)

        # non-atomic cleanups

        # at this point, the repo officially does not exist any more, the
        # roster file does not matter, cleanup workdirs if any.
        remove_all_workdirs_bare(self.repo_workdirs_root(request.repository,
                                                         context))
        shutil.rmtree(trash_path)
        return RemoveRepositoryResponse()

    def CreateBundle(self, request: CreateBundleRequest,
                     context) -> CreateBundleResponse:
        repo = self.load_repo(request.repository, context).unfiltered()
        yield from self.gen_bundle_responses(CreateBundleResponse, repo,
                                             all=True)

    def gen_bundle_responses(self, response_class, repo, **bundle_opts):
        """Create bundle and generate gRPC responses"""
        # overrides makes sure that 1) phases info 2) obsmarkers are bundled
        overrides = {
            (b'experimental', b'bundle-phases'): True,
            (b'experimental', b'evolution.bundle-obsmarker'): True,
        }
        # also bundle the hidden changesets unless explicitely excluded
        bundle_opts.setdefault('hidden', True)
        with tempfile.NamedTemporaryFile(
                mode='wb+',
                buffering=WRITE_BUFFER_SIZE) as tmpf:
            try:
                with repo.ui.configoverride(overrides, b'CreateBundle'):
                    bundle(repo.ui, repo, pycompat.sysbytes(tmpf.name),
                           **bundle_opts)
            finally:
                gc.collect()
            tmpf.seek(0)
            while True:
                data = tmpf.read(WRITE_BUFFER_SIZE)
                if not data:
                    break
                yield response_class(data=data)

    def CreateBundleFromRefList(self, request: CreateBundleFromRefListRequest,
                                context) -> CreateBundleFromRefListResponse:
        # TODO Notes (probably for discussion only, before merging):
        # 1) Get it working for `git bundle create my.bundle master ^master~1`
        logger = LoggerAdapter(base_logger, context)
        first_req = next(request)
        repo_msg = first_req.repository
        if not (repo_msg.storage_name or repo_msg.relative_path):
            context.abort(StatusCode.INVALID_ARGUMENT, 'repository not set')

        repo = self.load_repo(first_req.repository, context)
        patterns = itertools.chain(
            first_req.patterns,
            (pat for req in request for pat in req.patterns))

        incl_shas, excl_shas = resolve_revspecs_positive_negative(
            repo, patterns, ignore_unknown=True)
        logger.info("CreateBundleFromRefList repo=%r "
                    "included nodes=%r excluded nodes=%r",
                    message.Logging(first_req.repository),
                    incl_shas, excl_shas)

        # For info, in `hg bundle` one of the option from ('all', 'base')
        # is required, otherwise hg assumes that dest has all the nodes present
        incl_opts = {}
        if incl_shas is ALL_CHANGESETS:
            if not excl_shas:
                # underlying bundle command ignores --base if --all
                # is specified, but accepts --base without --rev, meaning
                # exactly what we need
                incl_opts['all'] = True
        else:
            incl_opts['rev'] = incl_shas

        if not excl_shas:
            excl_shas = [node.nullrev]

        yield from self.gen_bundle_responses(CreateBundleFromRefListResponse,
                                             repo.unfiltered(),
                                             base=excl_shas,
                                             **incl_opts)

    def FetchBundle(self, request: FetchBundleRequest,
                    context) -> FetchBundleResponse:
        logger = LoggerAdapter(base_logger, context)

        def load_or_init_repo(req, context):
            grpc_repo = req.repository
            created, repo = self.load_or_create_repo(grpc_repo, context)
            if not created:
                logger.info("FetchBundle: no need to create repo %r",
                            message.Logging(grpc_repo))
            return repo

        try:
            with streaming_request_tempfile_extract(
                    request, context,
                    first_request_handler=load_or_init_repo
            ) as (repo, tmpf):
                unbundle(repo, tmpf.name)
        finally:
            gc.collect()
            return CreateRepositoryFromBundleResponse()

    def CreateRepositoryFromBundle(
            self, request: CreateRepositoryFromBundleRequest,
            context) -> CreateRepositoryFromBundleResponse:
        """Create repository from bundle.

        param `request`: is an iterator streaming sub-requests where first
        sub-request contains repository+data and subsequent sub-requests
        contains only data (i.e. the actual bundle sent in parts).
        """
        with streaming_request_tempfile_extract(
                request, context,
                first_request_handler=lambda rq, c: self.create_and_load_repo(
                    rq.repository, c),
        ) as (repo, tmpf):
            try:
                unbundle(repo, tmpf.name)
            except Exception as exc:
                # same cleanup as Gitaly does, which gives later attempts
                # perhaps with a better bundle, chances to succeed.
                shutil.rmtree(repo.root)
                context.abort(StatusCode.INTERNAL,
                              "error in bundle application: %r" % exc)
            finally:
                gc.collect()
        return CreateRepositoryFromBundleResponse()

    def FindLicense(self, request: FindLicenseRequest,
                    context) -> FindLicenseResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context)
        # we'll work on the head of the default GitLab branch
        changeset = gitlab_revision_changeset(repo, b'HEAD')
        detector_bin_path = os.fsdecode(
            repo.ui.config(b'hgitaly', b'license-detector-executable')
        )
        try:
            with self.working_dir(gl_repo=request.repository,
                                  repo=repo,
                                  context=context,
                                  changeset=changeset) as wd:
                detected = detect_license(detector_bin_path, wd.path, context)
        except WorkingDirectoryError as exc:
            logger.warning("Failed to acquire a working directory"
                           "at stage %r" % exc.args[0])
            detected = None

        if detected is None:
            return FindLicenseResponse()
        return FindLicenseResponse(
            license_short_name=detected['spdx_id'],
            license_name=detected.get('full_name'),
            license_path=detected.get('file'),
            license_url=detected.get('url'),
            license_nickname=detected.get('nickname'),
        )


def render_git_grep_matches(buf, ref, path, enum_lines):
    """Render a slice of lines as git grep does.

    :param enum_lines: iterable of pairs `(line_no, line)`
    """
    for lineno, line in enum_lines:
        buf.write(b'%s:%s\x00%d\x00%s\n' % (ref, path, lineno, line))


SPLITLINES_RX = re.compile(br'(.*?)(\r\n?|\n|\Z)', re.MULTILINE)


def grep_file(rx, data, context_width=2):
    """Iterator yielding matches the given regular expression with context

    This implementation avoids duplicating the data in memory.

    :param int context_width: number of lines before and after to include
      if possible (same as `grep -C`)
    :returns: pairs `(line_no, line)`
    """
    prev_match_line_no = None
    current_window = deque()  # current line and up to 2 lines before
    match_lines = []
    # According the the best ranked answer for
    # https://stackoverflow.com/questions/3054604,
    # this regexp-based splitting for lines can be a bit slower than
    # `iter(splitlines)` (things may have changes since then, though).
    # Yet, our biggest concern is to avoid exhausting HGitaly's RAM budget
    # if this happens to run on large files.
    # Unfortunately, we already have to load the entire data in RAM because
    # it is typically Mercurial file content, we don't want to do it once
    # more in the form of lines. In both cases a bytes string
    # is allocated for each line, but that is harder to prevent and they
    # should be deallocated at each iteration (no cycles).
    # Unless there is a bug in the current approach, it is not worth the
    # effort to try and further improve memory efficiency: implementing
    # in RHGitaly would be the way to go.
    for line_idx, line_m in enumerate(SPLITLINES_RX.finditer(data)):
        line = line_m.group(1)
        if not line:
            continue
        line_no = line_idx + 1
        current_window.append((line_no, line))
        if rx.search(line):
            if (
                    prev_match_line_no is None
                    or line_no - prev_match_line_no > 2 * context_width
            ):
                match_lines = list(current_window)
            else:
                match_lines.append((line_no, line))
            prev_match_line_no = line_no
        elif (
                prev_match_line_no is not None
                and line_no <= prev_match_line_no + context_width
        ):
            match_lines.append((line_no, line))
        elif match_lines:
            yield match_lines
            match_lines = []

        if len(current_window) > context_width:
            current_window.popleft()
    if match_lines:
        yield match_lines


def iter_sfbc_resps(match_data: BytesIO):
    """Yield SearchFilesByContentResponse messages for the given match_data.
    """
    value = match_data.getvalue()
    match_len = len(value)
    sent = 0  # actually rounded to the upper WRITE_BUFFER_SIZE
    while sent < match_len:
        yield SearchFilesByContentResponse(
            match_data=value[sent:sent + WRITE_BUFFER_SIZE])
        sent += WRITE_BUFFER_SIZE

    match_data.truncate(0)
    match_data.seek(0)
    yield SearchFilesByContentResponse(end_of_match=True)



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from grpc import StatusCode
import logging
import re

from mercurial import (
    error,
    scmutil,
)
from mercurial.context import changectx
from hgext3rd.heptapod.branch import get_default_gitlab_branch
from hgext3rd.heptapod.keep_around import (
    parse_keep_around_ref,
)
from heptapod.gitlab.change import (
    ZERO_SHA,
)
from heptapod.gitlab.branch import (
    gitlab_branch_from_ref,
)
from heptapod.gitlab.tag import (
    gitlab_tag_from_ref,
)
from .branch import (
    gitlab_branch_head
)
from .gitlab_ref import (
    gitlab_special_ref_target
)
from .tag import tagged_changeset

logger = logging.getLogger(__name__)

# 40 hex digits for SHA-1, 64 hex digits for SHA-2
_HASH_RX = r'^([0-9A-Fa-f]{40})|([0-9A-Fa-f]{64})$'
CHANGESET_HASH_BYTES_REGEXP = re.compile(_HASH_RX.encode('ascii'))
CHANGESET_HASH_STR_REGEXP = re.compile(_HASH_RX)

ZERO_SHA_STR = ZERO_SHA.decode('ascii')


class RevisionNotFound(LookupError):
    pass


def gitlab_revision_changeset(repo, revision):
    """Find the changeset for a given GitLab revision.

    In theory, a GitLab revision could be any Git valid revspec, that
    we'd had to translate into its Mercurial counterpart.

    At this point, we support changeset IDs in hex, GitLab branches, tags and
    ``HEAD`` (default GitLab branch).

    Obsolescence
    ------------
    Changeset IDs can return obsolete changesets (this is actually used in
    force push detection, or in Merge Request updates), while symbolic ones
    (branch, tags etc) *should* always return non obsolete changesets.

    :return: the changeset as a :class:`changectx` instance, or ``None``
             if not found.
    """
    if CHANGESET_HASH_BYTES_REGEXP.match(revision):
        try:
            ctx = repo.unfiltered()[revision]
        except error.RepoLookupError:
            return None  # voluntarily explicit

        return ctx if isinstance(ctx, changectx) else None

    if revision == b'HEAD':
        revision = get_default_gitlab_branch(repo)
        if revision is None:
            return None

    ka = parse_keep_around_ref(revision)
    if ka is not None:
        try:
            return repo.unfiltered()[ka]
        except error.RepoLookupError:
            logger.warning("Unresolvable keep-around %r", ka)
            return None

    # non ambigous ref forms: they should obviously take precedence
    gl_tag = gitlab_tag_from_ref(revision)
    if gl_tag is not None:
        return tagged_changeset(repo, gl_tag)

    gl_branch = gitlab_branch_from_ref(revision)
    if gl_branch is not None:
        # can return None
        return gitlab_branch_head(repo, gl_branch)

    # special ref
    ctx = gitlab_special_ref_target(repo.unfiltered(), revision)
    if ctx is not None:
        return ctx

    # direct GitLab tag name
    ctx = tagged_changeset(repo, revision)
    if ctx is not None:
        return ctx

    # direct GitLab branch name
    ctx = gitlab_branch_head(repo, revision)
    if ctx is not None:
        return ctx

    try:
        ctx = scmutil.revsingle(repo.unfiltered(), revision)
        # ctx can be, e.g., workingctx (see heptapod#717)
        # The null node is not to be filtered (it may be used in diffs)
        return ctx if isinstance(ctx, changectx) else None
    except error.RepoLookupError:
        return None  # voluntarily explicit


def gitlab_revision_hash(repo, revision):
    """Return the changeset hexadecimal hash for given GitLab revision

    This is the preferred method to use when handling lists of revisions
    because the contexts retured by func:`gitlab_revision_changeset` may be
    associated to different repository views (filtered or not), hence it
    can be a trap for caller code to use them in revsets.

    Besides the return type, a major difference with
    func:`gitlab_revision_changeset` is that it raises instead of returning
    ``None`` if the revision could not be resolved. This is also more
    handy in loops and comprehensions than checking for or filtering out
    ``None``.

    :returns: the hash, as an hexadecimal bytes string.
    :raises RevisionNotFound:
    """
    ctx = gitlab_revision_changeset(repo, revision)
    if ctx is None:
        raise RevisionNotFound(revision)
    return ctx.hex()


ALL_CHANGESETS = object()


def resolve_revspecs_positive_negative(repo, revisions,
                                       ignore_unknown=False):
    """Sort the given revision specs into positive/negative and resolve them.

    This method understand negative revspecs, such as ``^ref1`` or
    ``^12de34ad56be`` as well as positive ones, such as ``branch/default`` or
    ``12de34ad56be``.

    It also accepts the special ``ALL`` positive revspec,
    which triggers returning :const:`ALL_CHANGESETS` as the positive set
    and avoid all positive resolutions once it is encountered, as they would
    be useless.

    :param revisions: any iterable of bytes. Generators will be consumed.
    :param ignore_unknown: if ``False``, raise ``RevisionNotFound`` on the
       first encountered unknown revision. Otherwise, just ignore it.
    :returns: a pair ``(positive, negative)``. In general, these are
      :class:`set` instances, whose elements are hexadecimal node ids
      as bytes. As a special case, ``positive`` can be ``ALL_CHANGESETS``.

    Typically, the caller will want to interpret each sets with an inner ``OR``
    and connect them as ``AND NOT``. In terms of revset, if ``positive`` is
`   ``{pos1, pos2}`` and ``negative`` is ``{neg1, neg2}``, this translates as
    ``(pos1, pos2) % (neg1, neg2)``
    """
    resolved = {}
    positive = set()
    negative = set()
    for rev in revisions:
        logger.debug("resolve_revspecs_positive_negative repo %r, treating "
                     "revspec %r", repo.root, rev)
        if rev.startswith(b'^'):
            collection = negative
            rev = rev[1:]
        else:
            if rev == b'ALL':
                positive = ALL_CHANGESETS
            collection = positive

        if collection is ALL_CHANGESETS:
            # resolution is useless
            continue

        sha = resolved.get(rev)
        if sha is None:
            try:
                sha = gitlab_revision_hash(repo, rev)
            except RevisionNotFound:
                if ignore_unknown:
                    continue
                raise
        resolved[rev] = sha
        collection.add(sha)

    return positive, negative


def validate_oid(oid: str):
    """Check that the given commit_id is hexadecimal of correct length
    """
    return CHANGESET_HASH_STR_REGEXP.match(oid) is not None


def changeset_by_commit_id(repo, commit_id: str):
    """Return a changeset context from an exact full hexadecimal Node ID.

    :raises ValueError: if `commit_id` is not an hexadecimal number of correct
         length.
    :raises KeyError: if changeset not found.
    """
    if not validate_oid(commit_id):
        raise ValueError(commit_id)
    try:
        return repo[commit_id.encode('ascii')]
    except error.RepoLookupError:
        raise KeyError(commit_id)


def changeset_by_commit_id_abort(repo, commit_id: str, context):
    """Same as :func:`changeset_by_commit_id`, with some error treatment

    aborts gRPC context if ``commit_id`` is not an hexadecimal number of
    correct length.

    :return: the changeset context, or ``None`` if resolution failed
    """
    try:
        return changeset_by_commit_id(repo, commit_id)
    except KeyError:
        return None
    except ValueError:
        context.abort(StatusCode.INVALID_ARGUMENT,
                      f'cannot parse commit ID: "{commit_id}"')
